/*
Copyright (c) 2009, Joel Reardon
All rights reserved.

Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

    * Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
    * Neither the name of the University of Waterloo nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/


#include <math.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "spe.h"

spe_t* _spe;

gint intcmp(gconstpointer a, gconstpointer b) {
 	if ((int)a == (int)b) return 0;
	if ((int)a < (int)b) return -1;
	return 1;
}


int compare_ints( const void* a, const void* b ) {
   int* arg1 = (int*) a;
   int* arg2 = (int*) b;
   if( *arg1 < *arg2 ) return -1;
   else if( *arg1 == *arg2 ) return 0;
   else return 1;
 }        


/* returns the difference of two time vals in microseconds.
   Assumes the difference can fit in an int type
*/
int spe_time_diff(struct timeval* later, struct timeval* earlier) {
	int retval = later->tv_sec - earlier->tv_sec;
	retval *= 1000000;
	retval += later->tv_usec - earlier->tv_usec;
	return retval;
} //time_diff 
 
/* Copies the time from src into dst */
void spe_time_copy(struct timeval* dst, struct timeval* src) {
	dst->tv_usec = src->tv_usec;
	dst->tv_sec = src->tv_sec;

}

/// Adds the time in <i>src</i> into <i>dst</i>
void spe_time_add(struct timeval* dst, struct timeval* src) {
	dst->tv_usec += src->tv_usec;
	dst->tv_sec  += src->tv_sec;
	if (dst->tv_usec >= 1000000){
		dst->tv_usec %= 1000000;
		dst->tv_sec +=1;
	}
}

///fills <i>tv</i> with the time of day
void spe_gettimeofday(struct timeval* tv) {
	gettimeofday(tv, NULL);
}

///fills the time field in <i>var</i> with the time of day
void spe_gettime(spe_var_t * var) {
	spe_gettimeofday(&var->timing);
	var->is_valid = 1;
}

///returns the difference in time between the current time and the time field in <i>var</i> 
int spe_gettimediff(spe_var_t * var) {
	struct timeval tv;
	int retval;
	spe_gettimeofday(&tv);
	retval = spe_time_diff(&tv, &var->timing);
	var->is_valid = 0;
	return retval;
}

/// initializes a new spe context
spe_t * spe_init() {

	spe_t * spe_context  = malloc (sizeof(spe_t));
	spe_gettimeofday(&spe_context->init_time);
	spe_context->spe_families = g_hash_table_new_full(g_str_hash , g_str_equal, NULL, (void (*)(void*))spe_varfamily_delete);
//	spe_context->buf_mgr = spe_buffer_manager_new();
	spe_context->obs_mgr = spe_observer_manager_new();
	spe_context->dataset_size = 0;
	spe_context->collect_indexed_vars = 0;
	pthread_mutex_init(&spe_context->socket_lock, NULL);
	


  	return spe_context;

}

void spe_delete(spe_t * context) {
	g_hash_table_destroy(context->spe_families);
	
	pthread_mutex_init(&context->socket_lock, NULL);
//	spe_buffer_manager_delete(context->buf_mgr);
	free(context);
}


void spe_delete_full(spe_t * context) {
	spe_socket_thread_stop(context);
//	spe_maintainance_thread_stop(context);
 	pthread_join(context->socket_thread, NULL);
//	pthread_join(context->maintainance_thread, NULL);

	spe_delete(context);
}


spe_t* spe_init_full(int port, int dataset_size) {
	spe_t* spe = spe_init();
	//spe_maintainance_thread_start(spe, trim_size);
	spe_reinit_full(spe, port, dataset_size);
	return spe;
}

void
spe_reinit_full(spe_t* spe,int port, int dataset_size){
		
		
	spe_socket_thread_start(spe, port);
	spe->dataset_size = dataset_size;
//	spe->dataset_default_size = dataset_default_size;
	spe->alive = 1;
}

/*
void spe_maintainance_thread_start(spe_t* spe, int trim_size) {

	spe->maintainance_trim_size = trim_size;
	spe->maintainance_thread_keepalive = 1;
	pthread_create(&spe->maintainance_thread, NULL, spe_maintainance_thread, spe);
	
}

void spe_maintainance_thread_stop(spe_t* spe) {
	spe->maintainance_thread_keepalive = 0;

}**/

void spe_socket_thread_start(spe_t* spe, int port) {
	spe->socket_port = port;
	spe->socket_thread_keepalive = 1;
	pthread_create(&spe->socket_thread,NULL,(void*(*)(void*))spe_listener_thread,spe);
	
}

void spe_socket_thread_stop(spe_t* spe) {
	spe->socket_thread_keepalive = 0;

}




spe_varfamily_t* spe_varfamily_new(const char* name, int data_type, int dataset_size) {

	spe_varfamily_t* varfam = malloc(sizeof(spe_varfamily_t));
	varfam->name = malloc(sizeof(char)*(strlen(name)+1));
	strcpy(varfam->name, name);
	varfam->data = g_hash_table_new_full(g_direct_hash, g_direct_equal, NULL, spe_switch_var_delete);
	varfam->magic = SPE_VARFAMILY;
	varfam->data_type = data_type;
	varfam->dataset_size = dataset_size;

	return varfam;

}


void spe_switch_var_delete(void* var) {
	int magic = *(int*)var;
	switch (magic) {
		case SPE_VAR:
			spe_var_delete(var);
			break;
		case SPE_VARSTATE:
			spe_varstate_delete(var);
			break;
		default:
			printf("nodel\n");
			break;
	}
	return;

}

void spe_varfamily_delete(spe_varfamily_t* varfam) {
	g_hash_table_destroy(varfam->data);
	free(varfam->name);
	free(varfam);
}

//void spe_varfamily_delete_fe(void* key, void* value, void* data) {
	//g_hash_table_remove(((spe_t*)data)->spe_families, key)
//	spe_varfamily_delete(value);
	
//	spe_clear_data(data, key);
//}

spe_varfamily_t* spe_varfamily_add(spe_t * context, const char* str, int data_type, int dataset_size) {
	spe_varfamily_t * varfam = spe_varfamily_new(str,data_type, dataset_size);
	g_hash_table_insert(context->spe_families, varfam->name, varfam);
	return varfam;
}

spe_varstate_t* spe_varstate_new(int dataset_size) {
	spe_varstate_t * var = malloc(sizeof(spe_varstate_t));
	
	pthread_mutex_init(&var->lock, NULL);
	var->magic = SPE_VARSTATE;
	var->states = g_hash_table_new_full(g_direct_hash, g_direct_equal, NULL, (void(*)(void*))spe_varstate_data_delete);
	var->cur_state = -1;
	var->timing.tv_sec = 0;
	var->timing.tv_usec = 0;
	var->dataset_size = dataset_size;
	return var;
}

void spe_varstate_delete(spe_varstate_t* varstate) {
	g_hash_table_destroy(varstate->states);
	pthread_mutex_destroy(&varstate->lock);
	free(varstate);
}

spe_varstate_t* spe_varstate_add(spe_varfamily_t* varfam, void* p) {
	spe_varstate_t * var = spe_varstate_new(varfam->dataset_size);
	g_hash_table_insert(varfam->data, p, var);
	return var; 
}


void* spe_var_data_new(int dataset_size) {
	if (dataset_size == 0) {
		return spe_data_list_new();
	}
	else {
		return spe_data_array_new(dataset_size);
	}
}

spe_data_list_t *spe_data_list_new() {
	
	spe_data_list_t* dl = malloc(sizeof(spe_data_list_t));
	dl->results = NULL;
	dl->magic = SPE_DATA_LIST;
	return dl;
}

spe_data_array_t *spe_data_array_new(int dataset_size) {
	
	spe_data_array_t* da = malloc(sizeof(spe_data_array_t));
	da->results = malloc(sizeof(int)*dataset_size);
	da->magic = SPE_DATA_ARRAY;
	da->curpos = 0;
	da->size = 0;
	da->dataset_size = dataset_size;
	return da;
}

void spe_var_data_delete(void* data) {
	int magic = *(int*) data;
	if (magic == SPE_DATA_LIST) {
		spe_data_list_t * dl = (spe_data_list_t*) data;
		g_slist_free(dl->results);
		free(dl);
	}
	else if (magic == SPE_DATA_ARRAY) {
		spe_data_array_t * da = (spe_data_array_t*) data;
		free(da->results);
		free(da);
	}
}

spe_var_t* spe_var_new(int dataset_size) {
	spe_var_t * var = malloc(sizeof(spe_var_t));
	
	var->is_valid = 0;
	pthread_mutex_init(&var->lock, NULL);
	var->magic = SPE_VAR;
	
	var->data = spe_var_data_new(dataset_size);
	return var;
	
}



void spe_var_delete(spe_var_t* var) {
	pthread_mutex_destroy(&var->lock);
	spe_var_data_delete(var->data);
	free(var);
}


spe_var_t* spe_var_add(spe_varfamily_t* varfam, void* p ) {
	spe_var_t * var ;
	if (p) //HACK
		var = spe_var_new(128);
	else
		var = spe_var_new(varfam->dataset_size);
	g_hash_table_insert(varfam->data, p, var);
	return var; 
}

spe_var_t * spe_var_fetch_add(spe_t * context, const char* str, void* p) {
	void* res = g_hash_table_lookup(context->spe_families, str);
	spe_varfamily_t * varfam = NULL;
	spe_var_t * var = NULL;	
	
	if (!res) {
		varfam = spe_varfamily_add(context, str, SPE_VAR, context->dataset_size);
	}
	else {
		varfam = (spe_varfamily_t *) res;
	}

	res = g_hash_table_lookup(varfam->data, p);

	if (!res) {		
		var = spe_var_add(varfam, p);
	}
	else {
		var = (spe_var_t *) res;
	}
	return var;
}

spe_varstate_t * spe_varstate_fetch_add(spe_t * context, const char* str, void* p) {
	void* res = g_hash_table_lookup(context->spe_families, str);
	spe_varfamily_t * varfam = NULL;
	spe_varstate_t * var = NULL;	
	
	if (!res) {
		varfam = spe_varfamily_add(context, str, SPE_VARSTATE, context->dataset_size);
	}
	else {
		varfam = (spe_varfamily_t *) res;
	}
	res = g_hash_table_lookup(varfam->data, p);
	
	if (!res) {
		var = spe_varstate_add(varfam, p);
	}
	else {
		var = (spe_varstate_t *) res;
	}


	return var;
}

void spe_timing_start(spe_t * context, const char* str, void* p) {
	if (p && !context->collect_indexed_vars) return;
	spe_var_t * var = spe_var_fetch_add(context, str, p);
	pthread_mutex_lock(&var->lock);
	spe_gettime(var);
	pthread_mutex_unlock(&var->lock);
}

void* spe_var_fetch(spe_t * context, const gchar* str, void* p ) {
	void* res = g_hash_table_lookup(context->spe_families, str);
	spe_varfamily_t * varfam;
	if (!res) return NULL;
	varfam = (spe_varfamily_t *) res;

	res = g_hash_table_lookup(varfam->data, p);
	return res;
	
}

void spe_timing_stop(spe_t * context, const gchar* str, void* p) {
	int diff;	
	if (p && !context->collect_indexed_vars) return;
	void* res = spe_var_fetch(context, str,p);

	if (res == NULL) return;
	spe_var_t * var = (spe_var_t *) res;
	if (!var->is_valid) return;	
	
	pthread_mutex_lock(&var->lock);
	diff = spe_gettimediff(var);
	


	spe_data_add_impl(var->data, diff);
	pthread_mutex_unlock(&var->lock);
}


void spe_timing_period(spe_t * context, const gchar* str, void* p) {
	if (p && !context->collect_indexed_vars) return;
	spe_var_t * var = spe_var_fetch_add(context, str, p);

	pthread_mutex_lock(&var->lock);
	if (var->is_valid) {
		//spe_time_copy(var->timing, context->init_time);
		int diff = spe_gettimediff(var);
		spe_data_add_impl(var->data, diff);
	}
	spe_gettime(var);
	pthread_mutex_unlock(&var->lock);
}



void spe_data_add_impl( void * var, int val) {
	int magic = *(int*)var;
	if (magic == SPE_DATA_LIST) {
		spe_data_list_t * lvar = (spe_data_list_t*) var;
		lvar->results = g_slist_prepend(lvar->results, (void*)val);
	}
	else if (magic == SPE_DATA_ARRAY) {
		spe_data_array_t * avar = (spe_data_array_t*) var;
		if (avar->size < avar-> dataset_size) {
			avar->results[avar->size++] = val;
		}
		else {
			avar->curpos = rand() % avar->dataset_size;
			avar->results[avar->curpos] = val;
			++avar->size;
		} 
	}
}

void spe_data_add(spe_t * context, const gchar* str, void* p, int data) {
	if (p && !context->collect_indexed_vars) return;
	spe_var_t * var = spe_var_fetch_add(context, str, p);

	pthread_mutex_lock(&var->lock);
	spe_data_add_impl(var->data, data);
	pthread_mutex_unlock(&var->lock);	

}

int spe_state_change(spe_t * context, const gchar* str, void* p, int newstate ) {
	if (p && !context->collect_indexed_vars) return;
	void * res=NULL;
	spe_varstate_t * var = spe_varstate_fetch_add(context, str, p);
	if (*(int*)var != SPE_VARSTATE) return -1;
	if (var->cur_state == newstate) return 0;
	pthread_mutex_lock(&var->lock);

	res = g_hash_table_lookup(var->states, (void*)var->cur_state);
	if (res) {
		//state has been added ==> timing is valid
		struct timeval tv;
		int diff;
	
		spe_varstate_data_t * vsdata = (spe_varstate_data_t*) res;
		++vsdata->count;
		
		spe_gettimeofday(&tv);
		diff = spe_time_diff(&tv , &var->timing);
		spe_data_add_impl(vsdata->data, diff);
	}
	spe_gettimeofday(&var->timing);
	res = g_hash_table_lookup(var->states, (void*)newstate);
	if (res == NULL) {
		g_hash_table_insert(var->states, (void*)newstate, spe_varstate_data_new(newstate,context->dataset_size));
	}
	var->cur_state = newstate;

	pthread_mutex_unlock(&var->lock);	
	return 0;
}

spe_varstate_data_t * spe_varstate_data_new(int state, int dataset_size) {
	spe_varstate_data_t * vsdata ;
	
	vsdata = malloc(sizeof(spe_varstate_data_t));
	vsdata->data = spe_var_data_new(dataset_size);
	vsdata->magic = SPE_VARSTATE_DATA;
	vsdata->count = 0;
	vsdata->dataset_size = dataset_size;
	vsdata->state = state;
	return vsdata;
}

void spe_varstate_data_delete(spe_varstate_data_t * vsdata) {
	spe_var_data_delete(vsdata->data);
	free(vsdata);	
}


void spe_clear_data(spe_t *context, const gchar* str) {
	void* res = g_hash_table_lookup(context->spe_families, str);
	if (!res) return;
	spe_varfamily_t * varfam = (spe_varfamily_t*) res;
	int data_type = varfam->data_type;
	spe_varfamily_delete(varfam);
	varfam = spe_varfamily_new(str, data_type, context->dataset_size);
	g_hash_table_insert(context->spe_families, varfam->name, varfam);

}

double spe_get_stdev(void * data, double avr) {
	int magic = *(int*)data;
	if (magic == SPE_DATA_ARRAY) {
		spe_data_array_t* da = (spe_data_array_t*) data;
		return spe_get_stdev_array(da->results, spe_var_data_length(da), avr);
	}
	else if (magic == SPE_DATA_LIST) {
		spe_data_list_t* la = (spe_data_list_t*) data;
		return spe_get_stdev_list(la->results, avr);
	}
}


double spe_get_average(void * data) {
	int magic = *(int*)data;
	if (magic == SPE_DATA_ARRAY) {
		spe_data_array_t* da = (spe_data_array_t*) data;
		return spe_get_average_array(da->results, spe_var_data_length(da));
	}
	else if (magic == SPE_DATA_LIST) {
		spe_data_list_t* la = (spe_data_list_t*) data;
		return spe_get_average_list(la->results);
	}
}

void spe_var_write_cdf(void * data, FILE* file) {
	int magic = *(int*)data;
	if (magic == SPE_DATA_ARRAY) {
		spe_data_array_t* da = (spe_data_array_t*) data;
		spe_var_write_cdf_array(da->results, spe_var_data_length(da), file);
	}
	else if (magic == SPE_DATA_LIST) {
		spe_data_list_t* la = (spe_data_list_t*) data;
		spe_var_write_cdf_list(la->results, file);
	}
}

double spe_get_stdev_array(int* data, int n, double avr) {
	double ret=0.;
	int i=0;

	for (; i < n ; ++i) {
		ret += (avr -(int) data[i])*(avr - (int)data[i]);
	}
	ret = ret/(double)(i-1);
	ret = sqrt(ret);
	return ret;

}
double spe_get_average_array(int* data, int n) {
	long long sum=0l;
	double ret;
	int i=0;
	for (; i < n; ++i){
		sum += data[i];
	}
	ret = (double)sum/(double)i;
	return ret;
}

void spe_var_write_cdf_array(int* data, int n, FILE* file) {

	int i =0;
	int l =0;
	for (i = 0 ; i < n; ++i) {
		while ((i*100)/n >=l) {
			l++;
			fprintf(file, ", %d", data[i]);
		}
	}

}

double spe_get_stdev_list(GSList* list, double avr) {
	double ret=0.;
	int i=0;
	GSList * cur = list;
	while (cur) {
		ret += (avr -(int) cur->data)*(avr - (int)cur->data);
		cur = cur->next;
		++i;
	}
	ret = ret/(double)(i-1);
	ret = sqrt(ret);
	return ret;

}
double spe_get_average_list(GSList * list) {
	long long sum=0l;
	double ret;
	int i=0;
	GSList * cur = list;
	while (cur) {
		sum += (int)cur->data;
		cur = cur->next;
		++i;
	}
	
	ret = (double)sum/(double)i;
	return ret;
}

void spe_var_write_cdf_list(GSList * cur, FILE* file) {

	int n = g_slist_length(cur);
	int i =0;
	int l =0;
	while (cur) {
		while ((i*100)/n >=l) {
			l++;
			fprintf(file, ", %d", (int)cur->data);
		}
		cur = cur->next;
		++i;
	}
	
}

void spe_state_write(void* key, void* value, void* file) {
	
	if (*(int*)value != SPE_VARSTATE_DATA) return;
	FILE * fout = (FILE*) file;
	
	spe_varstate_data_t * vsdata = (spe_varstate_data_t *) value;
	fprintf(fout, ", %d:(%d, [0", vsdata->state, vsdata->count);
	spe_var_data_sort(vsdata->data, spe_var_data_length(vsdata->data));
	spe_var_write_cdf(vsdata->data, (FILE*) file);
	fprintf(file, "])");
	
}

int spe_var_data_length(void* var) {
	int magic = *(int*)var;
	if (magic == SPE_DATA_LIST) {
		return g_slist_length(((spe_data_list_t* )var)->results);
	} else if (magic == SPE_DATA_ARRAY) {
		spe_data_array_t * avar = (spe_data_array_t*)var;
		return (avar->size < avar->dataset_size) ? avar->size: avar->dataset_size;
	}
	
}

void spe_var_data_sort(void * var, int len) {
	int magic = *(int*)var;
	if (magic == SPE_DATA_LIST) {
		spe_data_list_t* lvar = (spe_data_list_t*) var;
		lvar->results = g_slist_sort(lvar->results, intcmp);
		
	} else if (magic == SPE_DATA_ARRAY) {
		spe_data_array_t* avar = (spe_data_array_t*) var;
		qsort(avar->results, len, sizeof(int),compare_ints);		
		
	}
	
}	

void spe_var_write( void* key, void* value, void* file) {
	FILE * fout = (FILE*) file;
	
	if (*(int*)value == SPE_VAR) {
		spe_var_t * var = (spe_var_t * )value;
		int len = spe_var_data_length(var->data);
		if (len < 10) return;
		spe_var_data_sort(var->data, len);
		if (key == NULL) { fprintf(file, "default: (%d, [0",len);}
		else {fprintf(file, "%p: (%d, [0",key, len);}
		spe_var_write_cdf(var->data, file);
		fprintf(file, "])\n");
		
	} else if (*(int*)value == SPE_VARSTATE) {
		spe_varstate_t * var = (spe_varstate_t * )value;
		if (key == NULL) { fprintf(file, "default: {");}
		else {fprintf(file, "%p: {",key);}
		g_hash_table_foreach(var->states, spe_state_write, file);
		fprintf(file, "}\n");

	}


}

void spe_varfam_write_entry( void* key, void* value, void* file) {
	FILE * fout = (FILE*) file;
	
	spe_varfamily_t* varfam = (spe_varfamily_t*) value;
	if (varfam->magic != SPE_VARFAMILY) return;
	fprintf(fout, "[%s]\n", (char*)key); 
	g_hash_table_foreach(varfam->data, spe_var_write, file);
	fprintf(fout, "\n");
	
}

char* _spe_named_type (int type) {
	if (type == SPE_VARSTATE){return "state";}
	else if (type == SPE_VAR){return "variable";}
	return "";
}

void spe_varfam_write_toc( void* key, void* value, void* file) {
	FILE * fout = (FILE*) file;
	spe_varfamily_t* varfam = (spe_varfamily_t*) value;
	if (varfam->magic != SPE_VARFAMILY) return;
	fprintf(fout, "%s: %s\n", (char*)key, _spe_named_type(varfam->data_type)); 
}


void spe_generate_report_file(spe_t * context, FILE* fout) {
	
	fprintf(fout, "[varfamilies]\n");
	g_hash_table_foreach(context->spe_families, spe_varfam_write_toc, (void*)fout);
	fprintf(fout, "\n");
	g_hash_table_foreach(context->spe_families, spe_varfam_write_entry, (void*)fout);
	g_hash_table_destroy(context->spe_families);
	context->spe_families = g_hash_table_new_full(g_str_hash , g_str_equal, NULL, (void(*)(void*))spe_varfamily_delete);
	
	
}

void spe_generate_report(spe_t * context, const gchar* str) {

	FILE* fout = fopen(str, "w");
	spe_generate_report_file(context, fout);
	fclose(fout);
}



//socket interaction

void spe_variable_stat(spe_t* spe, int val) {
	spe->collect_indexed_vars = val;
}
void spe_variable_view(spe_t * spe, GString * ret, void* var) {
	
}

void spe_variable_dump(spe_t * spe, FILE* fdump) {
	
}

void spe_variable_list(spe_t * spe, GString * ret) {

	
}


void spe_variable_set_state(spe_t* spe, void * ptr, int newstate) {

}
