/*
Copyright (c) 2009, Joel Reardon
All rights reserved.

Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

    * Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
    * Neither the name of the University of Waterloo nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/



#include "spe.h"
#include <string.h>
     

void spe_observer_manager_queue_add_impl(spe_observer_manager_t * om , spe_observer_t * obs);

int spe_observer_cmp(spe_observer_t* a, spe_observer_t* b) {
	
	return spe_time_diff(&a->next_time,&b->next_time);
	
}

spe_observer_t* spe_observer_register(spe_t* spe, char* name, void* observee, int family, int* metric) {
	struct timeval interval;
	interval.tv_sec = 0;
	interval.tv_usec = 0;
	return spe_observer_register_time(spe, name, observee, family, metric, &interval);

}


void spe_observer_begin(spe_t* spe, char* name) {


	if (spe->obs_mgr->logger->enabled) return;
	pthread_mutex_lock(&spe->obs_mgr->logger->lock);
	spe->obs_mgr->logger->f = fopen(name, "w");
	spe->obs_mgr->logger->enabled = 1;
	pthread_mutex_unlock(&spe->obs_mgr->logger->lock);		

}


void spe_observer_end(spe_t* spe) {
	if (!spe->obs_mgr->logger->enabled) return;
	pthread_mutex_lock(&spe->obs_mgr->logger->lock);
	fflush(spe->obs_mgr->logger->f);
	fclose(spe->obs_mgr->logger->f);
	spe->obs_mgr->logger->f = NULL;
	spe->obs_mgr->logger->enabled = 0;
	pthread_mutex_unlock(&spe->obs_mgr->logger->lock);		


}


spe_observer_t* 
spe_observer_register_time(spe_t* spe, char* name, void* observee, int family, 
						   int* metric, struct timeval * interval) {
	return spe_observer_register_full(spe, name, observee, family, metric, 
									  interval, NULL, NULL);

}
spe_observer_t * 
spe_observer_register_full(spe_t* spe, char* name, void* observee, int family, 
						   int* metric, struct timeval* interval, void* state, 
						   void* state_delete) {
	spe_observer_t * obs = malloc(sizeof(spe_observer_t));
	obs->observee = observee;
	obs->family = family; 
	obs->name = strdup(name);
	obs->interval.tv_sec = interval->tv_sec;
	obs->interval.tv_usec = interval->tv_usec;
	if (obs->interval.tv_sec < 0) obs->interval.tv_sec =0;
	if (obs->interval.tv_usec < 0) obs->interval.tv_usec =0;
	obs->metric = metric;
	obs->logger = NULL;
	obs->enabled = 0;
	obs->quota = -1;
	obs->interest = 0;
	obs->state = state;
	obs->state_delete = state_delete;
	pthread_mutex_init(&obs->lock, NULL);
	
	if (spe->obs_mgr) {
		obs->logger = spe->obs_mgr->logger;
		g_hash_table_insert(spe->obs_mgr->data, observee, obs);
	}
	return obs;
}

void spe_observee_deregister(spe_t* spe, void* observee) {
	spe_observer_t* ticket = (spe_observer_t*) g_hash_table_lookup(spe->obs_mgr->data, observee);
	spe_observer_deregister(spe, ticket);
}

void spe_observer_deregister(spe_t* spe, spe_observer_t* ticket) {
	if (!ticket) return;
	spe_observer_manager_queue_remove(spe->obs_mgr, ticket);
	g_hash_table_remove(spe->obs_mgr->data, ticket->observee);
	spe_observer_delete(ticket);
}

void spe_observer_delete(spe_observer_t* obs) {
	pthread_mutex_destroy(&obs->lock);
	free(obs->name);
	
	if (obs->state && obs->state_delete) obs->state_delete(obs->state);
	free(obs);
}



void spe_observee_disable(spe_t* spe, void * observee) {
	spe_observer_t* obs = (spe_observer_t*) g_hash_table_lookup(spe->obs_mgr->data, observee);
	spe_observer_disable(spe->obs_mgr, obs);
}

void spe_observee_enable(spe_t* spe, void * observee, char* file) {
	spe_observer_t* obs = (spe_observer_t*) g_hash_table_lookup(spe->obs_mgr->data, observee);
	spe_observer_enable(spe->obs_mgr, obs, file);
}

void spe_observer_disable_fe(void* key, void* value, void* ptr) {
	spe_observer_t* obs = (spe_observer_t*) value;
	if (obs && obs->enabled) spe_observer_disable(ptr, obs);			
}

void spe_observer_disable_impl(spe_observer_manager_t* obs_mgr, spe_observer_t * ticket) {
	if (!ticket) return;
	if (ticket->enabled == 0) return;
	spe_observation_make(obs_mgr, ticket, SPE_OBSERVER_END);
	ticket->enabled = 0;
}

void spe_observer_disable_from_queue(spe_observer_manager_t* obs_mgr, spe_observer_t * ticket) {
	spe_observer_disable_impl(obs_mgr, ticket);
}

void spe_observer_disable(spe_observer_manager_t* obs_mgr, spe_observer_t * ticket) {
	spe_observer_disable_impl(obs_mgr, ticket);
	spe_observer_manager_queue_remove(obs_mgr, ticket);
}

void spe_observer_enable(spe_observer_manager_t* obs_mgr, spe_observer_t * ticket, char* file) {
	if (!ticket) return;
	if (ticket->enabled == 1) return;
//	if (ticket->interest == 0) return;
	ticket->enabled = 1;
	spe_observation_make(obs_mgr, ticket, SPE_OBSERVER_BEGIN);
	spe_observer_manager_queue_add(obs_mgr, ticket);
	
}


void spe_observee_make(spe_t* spe, void* observee, int obs_type){
	spe_observation_make(spe->obs_mgr, (spe_observer_t*) g_hash_table_lookup(spe->obs_mgr->data, observee), obs_type);
	
}

void spe_observation_make(spe_observer_manager_t* om, spe_observer_t* ticket, int obs_type) {
	if (!ticket) return;
	++(ticket->interest);

	if (ticket->enabled) {
		void (*fn)(int, void*, void**, spe_logger_t*);

		pthread_mutex_lock(&ticket->lock);
		*(void **)(&fn) = spe_fl_get(om->fl, ticket->family);
		if (fn) (*fn)(obs_type, ticket->observee, &ticket->state, ticket->logger);
		pthread_mutex_unlock(&ticket->lock);

	}

}


void spe_observer_next_observation_time(spe_observer_t* obs) {
	struct timeval now;
	spe_gettimeofday(&now);
	
	spe_time_copy(&obs->next_time, &now);
	spe_time_add(&obs->next_time, &obs->interval);
}


//BEGIN observer mananger

spe_observer_manager_t * spe_observer_manager_new() {
	spe_observer_manager_t * om = malloc(sizeof(spe_observer_manager_t));
	om->data = g_hash_table_new_full(g_direct_hash, g_direct_equal, NULL, NULL);
	om->fl   = spe_fl_new();
	om->Q    = g_queue_new();
	om->Qlock = malloc(sizeof(pthread_mutex_t));
	om->Qcond = malloc(sizeof(pthread_cond_t));
	om->Qthread_quit = 0;

	om->logger = spe_logger_new();
	
		
	pthread_mutex_init(om->Qlock, NULL);
	pthread_cond_init(om->Qcond, NULL);
	pthread_create(&om->Qthread, NULL, (void*(*)(void*))spe_observer_manager_queue_thread, om);
	return om;
}

void spe_observer_manager_delete(spe_observer_manager_t * om) {
	g_hash_table_foreach(om->data, spe_observer_disable_fe,NULL);
	om->Qthread_quit = 1;
	pthread_join(om->Qthread, NULL);
	pthread_cond_destroy(om->Qcond);
	pthread_mutex_destroy(om->Qlock);
	g_hash_table_destroy(om->data);
	g_queue_free(om->Q);
	spe_fl_delete(om->fl);
	spe_logger_delete(om->logger);
	free(om->Qlock);
	free(om);
}



void spe_observer_manager_queue_remove(spe_observer_manager_t * om , spe_observer_t * obs) {
	pthread_mutex_lock(om->Qlock);
	g_queue_remove(om->Q, obs);
	pthread_mutex_unlock(om->Qlock);
}
void spe_observer_manager_queue_add(spe_observer_manager_t * om , spe_observer_t * obs) {
	if (obs->interval.tv_sec == 0 && obs->interval.tv_usec == 0) return;
	pthread_mutex_lock(om->Qlock);
	spe_observer_manager_queue_add_impl(om, obs);
	pthread_cond_signal(om->Qcond);
	pthread_mutex_unlock(om->Qlock);
}

void spe_observer_manager_queue_add_impl(spe_observer_manager_t * om , spe_observer_t * obs) {
	spe_observer_next_observation_time(obs);
	g_queue_insert_sorted(om->Q, obs, (int(*)(const void*, const void*, void*))spe_observer_cmp, NULL);
}

void * spe_observer_manager_queue_thread ( spe_observer_manager_t * om) {
	struct timeval tv_now;
	struct timeval *tv_first;
	struct timespec ts;
	pthread_mutex_lock(om->Qlock);
	int i = 0;
	while (!om->Qthread_quit) {

		if (!g_queue_get_length(om->Q)) {
			struct timespec ts;
			ts.tv_sec = 1;
			ts.tv_nsec = 0;
			pthread_cond_timedwait(om->Qcond, om->Qlock, &ts);
		}
		else {
			tv_first = spe_observer_manager_queue_first_time(om->Q);
			spe_gettimeofday(&tv_now);
			ts.tv_sec = tv_first->tv_sec - tv_now.tv_sec;
			ts.tv_nsec = 1000*(tv_first->tv_usec - tv_now.tv_usec)+1000000;
			if (ts.tv_nsec < 0) {
				ts.tv_nsec += 1000000000;
				--(ts.tv_sec);
			}
			if (ts.tv_sec < 0) {
				spe_observer_manager_queue_observation_make(om);
			}
			else {
				pthread_cond_timedwait(om->Qcond, om->Qlock, &ts);
			}
		}
		++i;
		if (i == 1000000) {
			//starvation prevention
			i = 0;
			pthread_mutex_unlock(om->Qlock);
			sleep(1);
			pthread_mutex_lock(om->Qlock);
		}
	}
	pthread_mutex_unlock(om->Qlock);
	pthread_exit(NULL);
}


void spe_observer_manager_queue_observation_make ( spe_observer_manager_t* om) {
	spe_observer_t* os = g_queue_pop_head(om->Q);
	spe_observation_make(om, os, SPE_OBSERVER_OBSERVE);
	pthread_mutex_lock(&os->lock);
	
	if (os->quota == 1) 
			os->quota = -1;
	else 
			spe_observer_manager_queue_add_impl(om, os);
	
	if (os->quota > 0) os->quota -= 1;

	pthread_mutex_unlock(&os->lock);
	if (os->quota == -1) spe_observer_disable_from_queue(om, os);
}

struct timeval* spe_observer_manager_queue_first_time(GQueue * Q) {
	spe_observer_t * first = (spe_observer_t*) g_queue_peek_head(Q);
	return &(first->next_time);
}

//END observer manager



//BEGIN observer interaction


void spe_observer_list_fe(void* key, void* data, void* parm){
	spe_observer_t* obs = (spe_observer_t*) data;
	int interest = obs->interest;
	obs->interest = 0;
	GString * gs = (GString *) parm;
	g_string_append_printf(gs, "<observer id=\"%p\" name=\"%s\" family=\"%d\" enabled=\"%d\" quota=\"%d\" metric=\"%d\" interval_s=\"%d\" interval_us=\"%d\" interest=\"%d\">\n", 
						   key, obs->name, obs->family, obs->enabled, obs->quota, obs->metric? *(obs->metric):0, 
						   (int)obs->interval.tv_sec, (int)obs->interval.tv_usec, interest);
	

}

void spe_observer_list(spe_t * spe, GString* gs){
	g_string_append_printf(gs, "<observers size=\"%d\">", g_hash_table_size(spe->obs_mgr->data));
	g_hash_table_foreach(spe->obs_mgr->data, spe_observer_list_fe, gs);
	gs =g_string_append(gs, "</observers>\n");
}



void spe_observer_view(spe_t * spe, GString* gs, void* key) {
	spe_observer_list_fe(key, g_hash_table_lookup(spe->obs_mgr->data, key), (void*)gs);
}


void spe_observer_dump(spe_t * spe, FILE* fdump) {
	spe_generate_report_file(spe, fdump);
	return;
}

void spe_observer_set_state_all_fe(void* ptr, spe_observer_t* ticket, set_state_t* state_data) {
	if (!ticket) return;
	if (ticket->family != state_data->family) return;
	if (ticket->enabled == state_data->newstate) return; 
	if (state_data->newstate) {
		pthread_mutex_lock(&ticket->lock);
		ticket->quota = state_data->quota;
		spe_time_copy(&ticket->interval, &state_data->tv);
		pthread_mutex_unlock(&ticket->lock);
		spe_observer_enable(state_data->spe->obs_mgr, ticket, state_data->filename);
	}
	else  {
		spe_observer_disable(state_data->spe->obs_mgr, ticket);
			pthread_mutex_lock(&ticket->lock);
		ticket->quota = state_data->quota;
		spe_time_copy(&ticket->interval, &state_data->tv);	
		pthread_mutex_unlock(&ticket->lock);
	}
		
}

void spe_observer_set_state_all(spe_t * spe, int family, int newstate, int quota, struct timeval *tv, char* filename) {
		set_state_t state_data;
		state_data.family = family;
		state_data.newstate = newstate;
		state_data.quota = quota;
		state_data.tv.tv_sec = tv->tv_sec;
		state_data.tv.tv_usec = tv->tv_usec;
		state_data.filename = filename;
		state_data.spe = spe;
		g_hash_table_foreach(spe->obs_mgr->data, (void(*)(void*, void*, void*))spe_observer_set_state_all_fe, &state_data);
		
}

void spe_observer_set_state(spe_t * spe, void* ptr, int newstate, int quota, struct timeval *tv, char* filename) {
	spe_observer_t* ticket = (spe_observer_t*) g_hash_table_lookup(spe->obs_mgr->data, ptr);
	if (!ticket) return;
	if (newstate) {
		spe_observer_enable(spe->obs_mgr, ticket, filename);
		ticket->quota = quota;
		spe_time_copy(&ticket->interval, tv);	
	}
	else  {
		spe_observer_disable(spe->obs_mgr, ticket);
		ticket->quota = quota;
		spe_time_copy(&ticket->interval, tv);	
	}
}


//END observer interaction
